import traceback


class RequestLoggerMiddleware(object):

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        _body_to_log = request.body
        response = None
        error = None
        try:
            response = self.get_response(request)
        except BaseException as e:
            error = e
        finally:
            print("-------------------------------------")
            print('Request: ')
            print('    path    :', request.path)
            print('    method  :', request.method)
            print('    AUTH    :', request.META.get('HTTP_AUTHORIZATION'))
            print('    GET     :', request.GET)
            print('    POST    :', request.POST)
            if _body_to_log:
                print('    BODY    :', _body_to_log.decode('utf-8', errors='ignore').replace('\n', '                ')[:200])
            if response:
                print('Response:')
                print('    status  :', response.status_code)
                print('    headers :')
                for h in response.items():
                    print('             ', h)
                if response.content:
                    print('    content :', response.content.decode('utf-8', errors='ignore').replace('\n', '                ')[:200])

                return response

            if error:
                print('Error:', traceback.format_exc().replace('\n', '                '))



# section 4.1.3 of OAUTH 2.0
from django.conf import settings
from django.http import JsonResponse
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from ratelimit.mixins import RatelimitMixin

from openid_connect_op.decorators import access_token_required


class UserInfoView(RatelimitMixin, View):
    ratelimit_key = 'ip'
    ratelimit_rate = '10/m'
    ratelimit_block = True
    ratelimit_method = 'ALL'

    @method_decorator(csrf_exempt)
    @method_decorator(access_token_required())
    def dispatch(self, request, *args, **kwargs):
        token_data = request.openid_access_token.token_data
        claims = token_data['claims']
        scopes = token_data['scope']

        scope_param = request.GET.get('scope', request.POST.get('scope', None))
        if scope_param:
            scopes = set(scope_param.split()).intersection(set(scopes))

        claim_values = settings.OPENID_USERINFO_PROVIDERS.get_claims(request.openid_access_token,
                                                                    scopes, claims)
        return JsonResponse(claim_values)



from urllib.parse import quote

from django.http.response import HttpResponseForbidden, HttpResponseRedirect
from django.views import View
from ratelimit.mixins import RatelimitMixin

from django.conf import settings

from openid_connect_op.models import OpenIDToken


class LogoutRequestView(RatelimitMixin, View):
    ratelimit_key = 'ip'
    ratelimit_rate = '10/m'
    ratelimit_block = True
    ratelimit_method = 'ALL'

    def dispatch(self, request, *args, **kwargs):
        id_token_hint = request.GET.get('id_token_hint', None)
        post_logout_redirect_uri = request.GET.get('post_logout_redirect_uri', '/')
        state = request.GET.get('state', None)

        if not id_token_hint:
            return HttpResponseForbidden('Must supply id_token_hint parameter')

        # remove all tokens associated with the user
        token = OpenIDToken.objects.filter(token_hash=OpenIDToken.get_token_hash(id_token_hint)).first()
        if not token:
            return HttpResponseForbidden('Invalid token')

        root_token = token.root_token
        root_token.related_tokens.all().delete()
        root_token.delete()

        # create next uri
        next_url = post_logout_redirect_uri
        if state:
            if '?' in next_url:
                next_url += '&'
            else:
                next_url += '?'
            next_url += 'state=' + quote(state)

        # and call logout on this server to remove any session we might have with the user
        return HttpResponseRedirect(settings.LOGOUT_URL + '?next=' + quote(next_url))



from urllib.parse import splitquery

from django.conf import settings
from django.contrib.auth import logout
from django.http import HttpResponseRedirect
from django.http.response import HttpResponseBadRequest
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.http import urlencode
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from openid_connect_op.models import OpenIDClient, OpenIDKey
from openid_connect_op.views.parameters import AuthenticationParameters
from . import OAuthRequestMixin
from .errors import OAuthError


class AuthenticationRequestView(OAuthRequestMixin, View):
    # noinspection PyAttributeOutsideInit
    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        if request.method not in ('GET', 'POST'):
            return HttpResponseBadRequest('Only GET or POST are supported on OpenID endpoint')

        try:
            self.unpack_auth_parameters(request)

            self.validate_max_age()

            client = self._get_client_or_raise_exception()

            if self.should_login(request):
                return self.authenticate(request)

            if self.should_request_user_consent(request, client):
                return self.request_user_consent(request, client)

            if 'code' in self.request_parameters.response_type:
                self.request_parameters.username = request.user.username
                return self.oauth_send_answer(request, {
                    'code': self.request_parameters.pack(ttl=60, prefix=b'AUTH',
                                                         key=OpenIDClient.self_instance().get_key(OpenIDKey.AES_KEY)
                                                         )
                })
            else:
                raise OAuthError(error='parameter_not_supported',
                                 error_description='Only "code" value of the "response_type" '
                                                   'is supported by this server')

        except OAuthError as err:
            return self.oauth_send_answer(request, {
                'error': err.error,
                'error_description': err.error_description
            })

    def should_request_user_consent(self, request, client):
        return not client.has_user_agreement(request.user) or 'consent' in self.request_parameters.prompt

    def should_login(self, request):
        if request.user.is_anonymous or 'login' in self.request_parameters.prompt:
            if 'none' in self.request_parameters.prompt:
                raise OAuthError(error='login_required',
                                 error_description='No prompt requested but user is not logged in')
            return True
        return False

    def _get_client_or_raise_exception(self):
        try:
            client = OpenIDClient.objects.get(client_id=self.request_parameters.client_id)
        except OpenIDClient.DoesNotExist:
            raise OAuthError(error='unauthorized_client',
                             error_description='The client is unauthorized to get authentication token')
        if not client.check_redirect_url(self.request_parameters.redirect_uri):
            raise OAuthError(error='unauthorized_client',
                             error_description='The client is unauthorized to get authentication token '
                                               'as the redirect_uri does not match')
        return client

    def validate_max_age(self):
        if self.request_parameters.max_age:
            raise OAuthError(error='parameter_not_supported',
                             error_description='Parameter max_age is currently not supported by the server')

    def unpack_auth_parameters(self, request):
        try:
            if 'authp' in request.GET:
                self.request_parameters = AuthenticationParameters.unpack(
                    request.GET['authp'].encode('ASCII'), key=OpenIDClient.self_instance().get_key(OpenIDKey.AES_KEY))
            else:
                return self.parse_request_parameters(request, AuthenticationParameters)
        except AttributeError as e:
            raise OAuthError(error='invalid_request_uri', error_description=str(e))

    def authenticate(self, request):
        # if user is already logged in, a re-logging was requested, so forget the user
        if not request.user.is_anonymous:
            logout(request)

        this_url_with_encrypted_params = self.get_this_url_with_params(request)
        return HttpResponseRedirect(
            settings.LOGIN_URL + '?' + urlencode(
                {
                    'next': this_url_with_encrypted_params
                }
            )
        )

    def request_user_consent(self, request, client):
        this_url_with_encrypted_params = self.get_this_url_with_params(request)
        return HttpResponseRedirect(
            reverse(settings.OPENID_USER_CONSENT_VIEW, kwargs=dict(client_id=client.id)) + '?' + urlencode(
                {
                    'next': this_url_with_encrypted_params
                }
            )
        )

    def get_this_url_with_params(self, request):
        server, query = splitquery(request.build_absolute_uri())
        params = {
            'authp': self.request_parameters.pack(key=OpenIDClient.self_instance().get_key(OpenIDKey.AES_KEY))
        }
        return server + '?' + urlencode(params)



import os

SECRET_KEY = 'fake-key'
INSTALLED_APPS = [
    'modeltranslation',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'django_bleach',

    'openid_connect_op',
    'tests',
]

ROOT_URLCONF = 'tests.urls'
USE_TZ = True
USE_I18N = True

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'db.sqlite3',
    }
}

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'tests.middleware.RequestLoggerMiddleware'
]

DEBUG = True

STATIC_URL = '/static/'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]


# OpenID server

APPEND_SLASH = False

# test client
RUNNING_TEST_SERVER_CLIENT = False
if 'TEST_SERVER_CLIENT' in os.environ:
    RUNNING_TEST_SERVER_CLIENT = True

    print("Running in test server with python-social-auth")

    AUTHENTICATION_BACKENDS = (
        'tests.backends.ConfiguredOpenIdConnectAuth',
        'django.contrib.auth.backends.ModelBackend',
    )

    INSTALLED_APPS += [
        'social_django'
    ]

    # make sure this is the same url as the one used for runserver
    SERVER_URL = 'http://127.0.0.1:8000'

    KEY = 'openid_test_client_id'
    SECRET = 'openid_test_client_secret'


LOGIN_URL = '/django/login/'
LOGOUT_URL = '/django/logout/'

LANGUAGES = (
    ('en', 'English'),
    ('cs', 'Czech')
)



import json

import pytest
from django.contrib.auth.models import User
from django.core.management import call_command

from openid_connect_op.models import OpenIDToken, OpenIDClient


@pytest.mark.django_db
class TestClientRegistrationRequest:
    @pytest.fixture(autouse=True)
    def init_jwk(self):
        call_command('create_jwt_keys')

    def test_client_registration_bad_token(self, client):
        resp = client.post(
            '/openid/register',
            json.dumps({}),
            content_type='application/json',
            HTTP_AUTHORIZATION=self.format_auth('abc'))
        assert resp.status_code == 403

    def test_client_registration_bad_token2(self, client, openid_client, user):
        token = OpenIDToken.create_token(openid_client, OpenIDToken.TOKEN_TYPE_ACCESS_BEARER_TOKEN,
                                         {}, 1e9, user)[0]
        resp = client.post(
            '/openid/register',
            json.dumps({}),
            content_type='application/json',
            HTTP_AUTHORIZATION=self.format_auth(token))

        assert resp.status_code == 403

    def test_client_registration_no_redirect_uri(self, client, openid_client, user):
        token = OpenIDToken.create_token(openid_client, OpenIDToken.TOKEN_TYPE_CLIENT_DYNAMIC_REGISTRATION,
                                         {}, 1e9, user)[0]
        resp = client.post(
            '/openid/register',
            json.dumps({}),
            content_type='application/json',
            HTTP_AUTHORIZATION=self.format_auth(token))

        assert resp.status_code == 400
        assert json.loads(resp.content.decode('utf-8')) == {
            'error': 'invalid_request_uri',
            'error_description': 'Required parameter with name "redirect_uris" is not present'
        }

    def test_client_registration_ok(self, client, openid_client, user):
        token = OpenIDToken.create_token(openid_client, OpenIDToken.TOKEN_TYPE_CLIENT_DYNAMIC_REGISTRATION,
                                         {}, 1e9, user)[0]
        resp = client.post(
            '/openid/register',
            json.dumps({
                'redirect_uris': [
                    'http://test.org/auth/complete'
                ],
                'another': 1
            }),
            content_type='application/json',
            HTTP_AUTHORIZATION=self.format_auth(token))

        assert resp.status_code == 201
        content = json.loads(resp.content.decode('utf-8'))
        assert 'client_id' in content
        assert 'client_secret' in content
        client_id = content.pop('client_id')
        client_secret = content.pop('client_secret')
        assert content == {
            'another': 1,
            'client_secret_expires_at': 0,
            'redirect_uris': ['http://test.org/auth/complete']
        }

        client = OpenIDClient.objects.get(client_id=client_id)
        assert client.check_client_secret(client_secret)


    @pytest.fixture()
    def openid_client(self):
        return OpenIDClient.objects.create(
            client_id='1',
            client_auth_type=OpenIDClient.CLIENT_AUTH_TYPE_BASIC,
        )

    @pytest.fixture
    def user(self):
        return User.objects.create(username='a')

    def format_auth(self, token):
        return 'Bearer ' + token



# section 4.1.3 of OAUTH 2.0
import json
try:
    import secrets
except ImportError:
    import openid_connect_op.utils.secrets_backport as secrets

from django.http import JsonResponse
from django.http.response import HttpResponseForbidden, HttpResponseBadRequest
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from ratelimit.mixins import RatelimitMixin

from openid_connect_op.decorators import access_token_required
from openid_connect_op.models import OpenIDToken, OpenIDClient
from openid_connect_op.views import OAuthRequestMixin, OAuthError
from openid_connect_op.views.parameters import DynamicClientRegistrationParameters


class DynamicClientRegistrationView(RatelimitMixin, OAuthRequestMixin, View):
    ratelimit_key = 'ip'
    ratelimit_rate = '10/m'
    ratelimit_block = True
    ratelimit_method = 'ALL'

    @method_decorator(csrf_exempt)
    @method_decorator(access_token_required(disabled_settings='OPENID_DYNAMIC_CLIENT_REGISTRATION_DISABLE_ACCESS_TOKEN'))
    def dispatch(self, request, *args, **kwargs):
        if request.openid_access_token and \
                        request.openid_access_token.token_type != OpenIDToken.TOKEN_TYPE_CLIENT_DYNAMIC_REGISTRATION:
            return HttpResponseForbidden('Can not use this token to access dynamic registration view')

        if request.method != 'POST':
            return HttpResponseBadRequest('Only POST is allowed')

        try:
            self.parse_request_parameters(request, DynamicClientRegistrationParameters)

            client_id = secrets.token_urlsafe(32)
            client_secret = secrets.token_urlsafe(32)

            client = OpenIDClient.objects.create(
                client_id=client_id,
                redirect_uris='\n'.join(self.request_parameters.redirect_uris),
                client_auth_type=OpenIDClient.CLIENT_AUTH_TYPE_BASIC,
                client_name=self.request_parameters.client_name
            )
            client.set_client_secret(client_secret)
            client.save()

            resp = json.loads(request.body.decode('utf-8'))
            resp.update({
                "client_id": client_id,
                "client_secret": client_secret,
                "client_secret_expires_at": 0,
            })

            return JsonResponse(resp, status=201)

        except OAuthError as err:
            return self.oauth_send_answer(request, {
                'error': err.error,
                'error_description': err.error_description
            })



import pytest

from openid_connect_op.utils.params import Parameters, ParameterType


class Params(Parameters):
    parameter_definitions = (
        ('a', Parameters.REQUIRED),
        ('b', Parameters.OPTIONAL),
        ('c', ParameterType(required=False, container_type=set)),
        ('d', ParameterType(required=False, container_type=list)),
        ('e', ParameterType(required=False, container_type=set, allowed_values=('a', 'b', 'c')))
    )


def test_parameter_parsing():
    p = Params({
        'a': 'abc',
        'b': 'def',
        'c': 'a ba',
        'd': 'ba a'
    })
    assert p.a == 'abc'
    assert p.b == 'def'
    assert p.c == {'a', 'ba'}
    assert p.d == ['ba', 'a']


def test_required_parameter():
    with pytest.raises(AttributeError) as e:
        Params({})
    assert str(e.value) == 'Required parameter with name "a" is not present'


def test_allowed_values():
    p = Params({
        'a': '123',
        'e': 'a c b'
    })
    assert p.a == '123'
    assert p.e == {'a', 'b', 'c'}

    with pytest.raises(AttributeError) as e:
        Params({
            'a': '123',
            'e': 'a c b d'
        })
    assert str(e.value) == 'Value "d" is not allowed for parameter e. Allowed values are "a", "b", "c"'


def test_pack_unpack():
    p = Params({
        'a': 'abc,',
        'b': 'def',
        'd': 'ba a'
    })
    assert isinstance(p.d, list)
    packed = p.pack(True, prefix=b'TOK', key=b'0123456789ABCDEF')
    unpacked_p = Params.unpack(packed, prefix=b'TOK', key=b'0123456789ABCDEF')
    assert p == unpacked_p
    assert str(p) == str(unpacked_p)



from django.contrib import admin
from modeltranslation.admin import TranslationAdmin, TranslationStackedInline

from openid_connect_op.models import OpenIDClient, OpenIDAgreement


class OpenIDAgreementAdmin(TranslationStackedInline):
    model = OpenIDAgreement


class OpenIDClientAdmin(TranslationAdmin):
    inlines = (OpenIDAgreementAdmin, )


admin.site.register(OpenIDClient, OpenIDClientAdmin)


from django.conf.urls import include, url
from django.contrib import admin

from tests.views import ConsentView, IndexView
urlpatterns = [
    url('^admin/', include(admin.site.urls)),
    url('^django/', include('django.contrib.auth.urls')),
    url('^', include('openid_connect_op.urls')),
    url('', include('social_django.urls', namespace='social')),
    # test views
    url('', include(
        (
            [
                url(r'^$', IndexView.as_view()),
            ],
            'test'
        ), namespace='test'))
]



import json
from urllib.parse import urlencode, parse_qs, splitquery

import pytest
from django.contrib.auth.models import User
from django.core.management import call_command

from openid_connect_op.models import OpenIDClient, OpenIDKey
from openid_connect_op.views.parameters import AuthenticationParameters


@pytest.mark.django_db
class TestAuthenticationRequest:
    @pytest.fixture(autouse=True)
    def init_jwk(self):
        call_command('create_jwt_keys')

    def test_no_redirect_uri_flow(self, client):
        resp = client.get('/openid/authorize')
        assert json.loads(resp.content.decode('utf-8')) == {
            "error": "invalid_request_uri",
            "error_description": "Required parameter with name \"redirect_uri\" is not present"}
        assert resp.status_code == 400

    def test_no_params_flow(self, client):
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': 'http://localhost:8000/complete/test/?state=1234'
        }))
        assert resp.status_code == 302
        assert resp.url.startswith('http://localhost:8000/complete/test/')
        self.check_query(resp, {
            'state': ['1234'],
            'error': ['invalid_request_uri'],
            'error_description': ['Required parameter with name "client_id" is not present']
        })

    @staticmethod
    def check_query(resp, expected):
        query = parse_qs(splitquery(resp.url)[1])
        assert query == expected

    def test_unknown_client_id_flow(self, client):
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': 'http://localhost:8000/complete/test/?state=1234',
            'client_id': '1'
        }))
        assert resp.status_code == 302
        self.check_query(resp, {
            'state': ['1234'],
            'error': ['invalid_request_uri'],
            'error_description': ['Required parameter with name "scope" is not present']
        })

    def test_no_response_type(self, client):
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': 'http://localhost:8000/complete/test/?state=1234',
            'client_id': '1',
            'scope': 'openid'
        }))
        assert resp.status_code == 302
        self.check_query(resp, {
            'state': ['1234'],
            'error': ['invalid_request_uri'],
            'error_description': ['Required parameter with name "response_type" is not present']
        })

    def test_unsupported_response_type(self, client):
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': 'http://localhost:8000/complete/test/?state=1234',
            'client_id': '1',
            'scope': 'openid',
            'response_type': 'unsupported_token'
        }))
        assert resp.status_code == 302
        self.check_query(resp, {
            'state': ['1234'],
            'error': ['invalid_request_uri'],
            'error_description': ['Value "unsupported_token" is not allowed for parameter '
                                  'response_type. Allowed values are "code", "id_token", "token"']
        })

    def test_unknown_client_id(self, client):
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': 'http://localhost:8000/complete/test/?state=1234',
            'client_id': 'unknown',
            'scope': 'openid',
            'response_type': 'code'
        }))
        assert resp.status_code == 302
        self.check_query(resp, {
            'state': ['1234'],
            'error': ['unauthorized_client'],
            'error_description': ['The client is unauthorized to get authentication token']
        })

    @pytest.fixture()
    def client_config(self):
        redirect_uri = 'http://localhost:8000/complete/test/?state=1234'
        return OpenIDClient.objects.create(
            client_id='test',
            redirect_uris=redirect_uri
        )

    def test_redirect_to_login(self, client, client_config):
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'client_id': 'test',
            'scope': 'openid',
            'response_type': 'code'
        }))
        self.check_is_redirect_to_login_page(resp)

    @staticmethod
    def check_is_redirect_to_login_page(resp):
        assert resp.status_code == 302

        # check that this is a url to the login server and there is a ?next=...
        login_server, login_query = splitquery(resp.url)
        assert login_server == '/django/login/'
        login_query = parse_qs(login_query)
        assert 'next' in login_query

        # parse the ?next=... and check that it goes back to the authorize endpoint with ?authp
        next_server, next_query = splitquery(login_query['next'][0])
        assert next_server == 'http://testserver/openid/authorize'
        next_query = parse_qs(next_query)
        assert 'authp' in next_query

        # check that ?authp param contains correctly encrypted value
        AuthenticationParameters.unpack(next_query['authp'][0].encode('utf-8'),
                                        key=OpenIDClient.self_instance().get_key(OpenIDKey.AES_KEY))

    @pytest.fixture
    def user(self):
        return User.objects.create(username='a')

    def test_logged_user(self, client, client_config, user):
        client.force_login(user)
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'client_id': 'test',
            'scope': 'openid',
            'response_type': 'code'
        }))

        self.check_response_contains_code(resp)

    @staticmethod
    def check_response_contains_code(resp):
        assert resp.status_code == 302

        redirect_server, redirect_query = splitquery(resp.url)
        assert redirect_server == 'http://localhost:8000/complete/test/'
        redirect_query = parse_qs(redirect_query)
        assert redirect_query['state'] == ['1234']
        assert 'code' in redirect_query

        # check that ?code param contains correctly encrypted value
        AuthenticationParameters.unpack(redirect_query['code'][0].encode('utf-8'), prefix=b'AUTH',
                                        key=OpenIDClient.self_instance().get_key(OpenIDKey.AES_KEY))

    def test_require_new_login(self, client, client_config, user):
        client.force_login(user)
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'client_id': 'test',
            'scope': 'openid',
            'response_type': 'code',
            'prompt': 'login'
        }))
        self.check_is_redirect_to_login_page(resp)

    def test_prompt_none_ok(self, client, client_config, user):
        client.force_login(user)
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'client_id': 'test',
            'scope': 'openid',
            'response_type': 'code',
            'prompt': 'none'
        }))

        self.check_response_contains_code(resp)

    def test_prompt_none_failure(self, client, client_config, user):
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'client_id': 'test',
            'scope': 'openid',
            'response_type': 'code',
            'prompt': 'none'
        }))
        assert resp.status_code == 302
        self.check_query(resp, {
            'state': ['1234'],
            'error': ['login_required'],
            'error_description': ['No prompt requested but user is not logged in']
        })

    def test_require_user_consent(self, client, client_config, user):
        client.force_login(user)
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'client_id': 'test',
            'scope': 'openid',
            'response_type': 'code',
            'prompt': 'consent'
        }))
        self.check_is_redirect_to_consent_page(client_config, resp)

    @staticmethod
    def check_is_redirect_to_consent_page(client_config, resp):
        assert resp.status_code == 302

        # check that this is a url to the login server and there is a ?next=...
        consent_server, consent_query = splitquery(resp.url)
        assert consent_server == '/openid/consent/%s/' % client_config.id
        consent_query = parse_qs(consent_query)
        assert 'next' in consent_query

        # parse the ?next=... and check that it goes back to the authorize endpoint with ?authp
        next_server, next_query = splitquery(consent_query['next'][0])
        assert next_server == 'http://testserver/openid/authorize'
        next_query = parse_qs(next_query)
        assert 'authp' in next_query

        # check that ?authp param contains correctly encrypted value
        AuthenticationParameters.unpack(next_query['authp'][0].encode('utf-8'),
                                        key=OpenIDClient.self_instance().get_key(OpenIDKey.AES_KEY))



import base64
import json
from urllib.parse import splitquery, parse_qs, urlencode

import pytest
from django.contrib.auth.models import User
from django.core.management import call_command

from openid_connect_op.models import OpenIDClient


@pytest.mark.django_db
class TestUserInfo:

    @pytest.fixture(autouse=True)
    def init_jwk(self):
        call_command('create_jwt_keys')

    @pytest.fixture
    def user(self):
        return User.objects.create(
            username='a',
            first_name='A',
            last_name='B',
            email='a@b.com')

    @pytest.fixture()
    def client_config(self):
        redirect_uri = 'http://localhost:8000/complete/test/?state=1234'
        ret = OpenIDClient.objects.create(
            client_id='test',
            redirect_uris=redirect_uri,
            client_auth_type=OpenIDClient.CLIENT_AUTH_TYPE_BASIC,
        )
        ret.set_client_secret('b')
        ret.save()
        return ret

    @staticmethod
    def get_access_code(client, client_config, user, extra_scopes, extra_claims):
        client.force_login(user)
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'client_id': 'test',
            'scope': '%s %s' % ('openid', ' '.join(extra_scopes)),
            'claims': ' '.join(extra_claims),
            'response_type': 'code'
        }))
        assert resp.status_code == 302
        redirect_server, redirect_query = splitquery(resp.url)
        assert redirect_server == 'http://localhost:8000/complete/test/'
        redirect_query = parse_qs(redirect_query)
        assert redirect_query['state'] == ['1234']
        assert 'code' in redirect_query
        code = redirect_query['code']

        resp = client.get('/openid/token?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'grant_type': 'authorization_code',
            'code': code,
        }), HTTP_AUTHORIZATION='Basic ' + base64.b64encode('test:b'.encode('utf-8')).decode('ascii'))

        data = json.loads(resp.content.decode('utf-8'))
        return data['access_token']

    def test_no_userinfo_claims(self, client, client_config, user):
        access_code = self.get_access_code(client, client_config, user, {}, {})

        resp = client.get('/openid/userinfo',
                          HTTP_AUTHORIZATION='Bearer ' + access_code)

        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {'sub': 'a'}

    def test_profile_scope(self, client, client_config, user):
        access_code = self.get_access_code(client, client_config, user, {'profile'}, {})

        resp = client.get('/openid/userinfo',
                          HTTP_AUTHORIZATION='Bearer ' + access_code)

        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {
            'sub': 'a',
            'name': 'A B',
            'family_name': 'B',
            'given_name': 'A',
            'preferred_username': 'a'
        }

    def test_profile_email_scope(self, client, client_config, user):
        access_code = self.get_access_code(client, client_config, user, {'profile', 'email'}, {})

        resp = client.get('/openid/userinfo',
                          HTTP_AUTHORIZATION='Bearer ' + access_code)

        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {
            'sub': 'a',
            'name': 'A B',
            'family_name': 'B',
            'given_name': 'A',
            'preferred_username': 'a',
            'email': 'a@b.com',
        }

    def test_email_scope(self, client, client_config, user):
        access_code = self.get_access_code(client, client_config, user, {'email'}, {})

        resp = client.get('/openid/userinfo',
                          HTTP_AUTHORIZATION='Bearer ' + access_code)

        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {
            'sub': 'a',
            'email': 'a@b.com',
        }

    def test_email_claim(self, client, client_config, user):
        access_code = self.get_access_code(client, client_config, user, {}, {'email'})

        resp = client.get('/openid/userinfo',
                          HTTP_AUTHORIZATION='Bearer ' + access_code)

        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {
            'sub': 'a',
            'email': 'a@b.com',
        }

    def test_name_claims(self, client, client_config, user):
        access_code = self.get_access_code(client, client_config, user, {}, {'family_name', 'given_name'})

        resp = client.get('/openid/userinfo',
                          HTTP_AUTHORIZATION='Bearer ' + access_code)

        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {
            'sub': 'a',
            'family_name': 'B',
            'given_name': 'A',
        }

    def test_allowed_claims_via_scope(self, client, client_config, user):
        client_config.allowed_scopes = ['profile']
        client_config.save()

        access_code = self.get_access_code(client, client_config, user, {'profile', 'email'}, {})

        resp = client.get('/openid/userinfo',
                          HTTP_AUTHORIZATION='Bearer ' + access_code)

        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {
            'sub': 'a',
            'name': 'A B',
            'family_name': 'B',
            'given_name': 'A',
            'preferred_username': 'a'
        }


    def test_allowed_claims_via_name(self, client, client_config, user):
        client_config.allowed_claims = ['name']
        client_config.save()

        access_code = self.get_access_code(client, client_config, user, {'profile', 'email'}, {})

        resp = client.get('/openid/userinfo',
                          HTTP_AUTHORIZATION='Bearer ' + access_code)

        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert data == {
            'sub': 'a',
            'name': 'A B',
        }


import base64
import json
from urllib.parse import urlencode, splitquery, parse_qs

import pytest
import time
from django.contrib.auth.models import User
from django.core.management import call_command

from openid_connect_op.models import OpenIDClient, OpenIDToken
from openid_connect_op.utils.jwt import JWTTools

BASIC_AUTH = 'Basic ' + base64.b64encode('test:b'.encode('utf-8')).decode('ascii')


@pytest.mark.django_db
class TestLogoutRequest:

    @pytest.fixture(autouse=True)
    def init_jwk(self):
        call_command('create_jwt_keys')

    @pytest.fixture
    def user(self):
        return User.objects.create(username='a')

    @pytest.fixture()
    def client_config(self):
        redirect_uri = 'http://localhost:8000/complete/test/?state=1234'
        ret = OpenIDClient.objects.create(
            client_id='test',
            redirect_uris=redirect_uri,
            client_auth_type=OpenIDClient.CLIENT_AUTH_TYPE_BASIC,
        )
        ret.set_client_secret('b')
        ret.save()
        return ret

    def test_user_logout(self, client, client_config, user, settings):
        code = self.get_authorization_code(client, client_config, user)
        resp = client.get('/openid/token?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'grant_type': 'authorization_code',
            'code': code,
        }), HTTP_AUTHORIZATION=BASIC_AUTH)

        access_token, id_token = self.check_token_response(settings, client_config, resp)

        resp = client.get('/openid/logout?' + urlencode({
            'id_token_hint': id_token,
            'post_logout_redirect_uri': 'https://example.com',
            'state': '2345',
        }))

        assert resp.status_code == 302
        assert resp.url == '/django/logout/?next=https%3A//example.com%3Fstate%3D2345'

    def test_no_id_token_logout(self, client, client_config, user, settings):
        resp = client.get('/openid/logout?' + urlencode({
            'post_logout_redirect_uri': 'https://example.com',
            'state': '2345',
        }))

        assert resp.status_code == 403
        assert resp.content == b'Must supply id_token_hint parameter'

    def test_bad_id_token_logout(self, client, client_config, user, settings):
        resp = client.get('/openid/logout?' + urlencode({
            'id_token_hint': '123456789',
            'post_logout_redirect_uri': 'https://example.com',
            'state': '2345',
        }))

        assert resp.status_code == 403
        assert resp.content == b'Invalid token'

    @staticmethod
    def check_token_response(settings, client_config, resp):
        assert resp.status_code == 200
        data = json.loads(resp.content.decode('utf-8'))
        assert 'access_token' in data
        assert data['token_type'] == 'Bearer'
        assert 'refresh_token' in data
        assert data['expires_in'] == settings.OPENID_DEFAULT_ACCESS_TOKEN_TTL
        assert 'id_token' in data
        database_at = OpenIDToken.objects.get(token_hash=OpenIDToken.get_token_hash(data['access_token']))
        assert database_at.user.username == 'a'
        assert database_at.client == client_config
        assert database_at.token_type == OpenIDToken.TOKEN_TYPE_ACCESS_BEARER_TOKEN
        database_rt = OpenIDToken.objects.get(token_hash=OpenIDToken.get_token_hash(data['refresh_token']))
        assert database_rt.user.username == 'a'
        assert database_rt.client == client_config
        assert database_rt.token_type == OpenIDToken.TOKEN_TYPE_REFRESH_TOKEN
        # validate id token
        header, payload = JWTTools.validate_jwt(data['id_token'])
        assert header['alg'] == 'RS256'
        assert header['typ'] == 'JWT'
        assert payload['exp'] == int(payload['exp'])
        assert payload['iat'] == int(payload['iat'])
        assert payload['aud'] == ['test']
        assert payload['sub'] == 'a'  # username
        assert payload['iss'] == 'http://testserver/'

        return data['access_token'], data['id_token']

    @staticmethod
    def get_authorization_code(client, client_config, user):
        client.force_login(user)
        resp = client.get('/openid/authorize?' + urlencode({
            'redirect_uri': client_config.redirect_uris,
            'client_id': 'test',
            'scope': 'openid',
            'response_type': 'code'
        }))
        assert resp.status_code == 302
        redirect_server, redirect_query = splitquery(resp.url)
        assert redirect_server == 'http://localhost:8000/complete/test/'
        redirect_query = parse_qs(redirect_query)
        assert redirect_query['state'] == ['1234']
        assert 'code' in redirect_query
        code = redirect_query['code']
        return code



import datetime
import hashlib
import logging

import re
from django.conf import settings
from django.db.models import Count, Sum
from jsonfield.fields import JSONField

from openid_connect_op.utils.crypto import CryptoTools

try:
    import secrets
except ImportError:
    import openid_connect_op.utils.secrets_backport as secrets

from urllib.parse import urlparse, splitquery, parse_qs

import jsonfield
from django.contrib.auth.hashers import get_hasher, check_password
from django.contrib.auth.models import User
from django.db import models
from django.utils import timezone

log = logging.getLogger(__file__)


class OpenIDClient(models.Model):
    """
    An abstract model that implements OpenID client configuration (client = someone who requests access token)
    """

    #
    # ID of the client, the client sends this id in the access request
    #
    client_id = models.CharField(max_length=128, unique=True)
    SELF_CLIENT_ID = '---self---'

    #
    # After logging in, browser is redirected to one of these URIs (separated by newline).
    # The actual redirection URI is sent by the client, OpenID server verifies that the URI
    # is among these configured ones. For detail of how this decision is made, see check_redirect_url
    # method.
    #
    redirect_uris = models.TextField(default='')

    CLIENT_AUTH_TYPE_BASIC = 'basic'
    CLIENT_AUTH_TYPE_POST = 'post'
    CLIENT_AUTH_TYPE_SECRET_JWT = 'sjwt'
    CLIENT_AUTH_TYPE_PRIVATE_KEY_JWT = 'pkjwt'
    CLIENT_AUTH_TYPE_NONE = 'none'
    CLIENT_AUTH_TYPE_INVALID = 'invalid'

    client_auth_type = models.CharField(max_length=8, choices=(
        (CLIENT_AUTH_TYPE_BASIC, 'Basic Authentication'),
        (CLIENT_AUTH_TYPE_POST, 'Authentication data in POST request'),
        (CLIENT_AUTH_TYPE_SECRET_JWT, 'JSON Web token with pre-shared secret'),
        (CLIENT_AUTH_TYPE_PRIVATE_KEY_JWT, 'JSON Web token with public/private key'),
        (CLIENT_AUTH_TYPE_NONE, 'No client authentication performed'),
    ))

    client_hashed_secret = models.CharField(max_length=128)

    client_name = models.CharField(max_length=128)

    allowed_scopes = JSONField(null=True, blank=True,
                               verbose_name='List of allowed scopes. If None, all scopes are allowed')

    allowed_claims = JSONField(null=True, blank=True,
                               verbose_name='List of allowed claims. If None, '
                                            'all claims from allowed scopes are returned')

    def set_client_secret(self, password):
        if password is None:
            raise AttributeError('Password can not be empty')
        hasher = get_hasher('default')
        salt = hasher.salt()
        self.client_hashed_secret = hasher.encode(password, salt)

    def check_client_secret(self, raw_password):
        # taken from User
        def setter(raw_password):
            self.set_client_secret(raw_password)
            self.save(update_fields=["client_hashed_password"])

        return check_password(raw_password, self.client_hashed_secret, setter)

    def has_user_agreement(self, user):
        """
        Checks if the user has approved sending his data (including, for example, roles, phone number etc.)
        to this client

        :param user:    django User
        :return:        True if user has approved sending the data (and client's usage policy), False otherwise
        """
        has_agreement = True

        for agreement in self.get_unsigned_agreements(user, auto_approve=True):
            if agreement.obligatory:
                has_agreement = False

        return has_agreement

    def get_unsigned_agreements(self, user, auto_approve=False):
        """
        Return all OpenIDAgreements for this client that user has not signed.
        :param user:            The user to check
        :param auto_approve:    If True, all approvals that can be signed automatically will be
        :return:                Generator of OpenIDAgreement
        """
        agreements_user_has_not_signed = \
            self.agreements.annotate(
                has_me=Sum(
                    models.Case(
                        models.When(user_agreements__user=user, then=1),
                        default=0,
                        output_field=models.IntegerField()))).filter(has_me=0)

        if not auto_approve:
            yield from agreements_user_has_not_signed

        for agreement in agreements_user_has_not_signed:
            # try to auto approve
            if agreement.can_auto_approve(user.username):
                OpenIDUserAgreement.objects.create(agreement=agreement, user=user, agreed_on=timezone.now())
            else:
                yield agreement

    def get_user_agreements(self, user):
        # return all the agreements that are applicable to the user. In the current version, that means all
        return self.agreements.all()

    def check_redirect_url(self, _redirect_uri):
        """
        Checks if the actual redirection uri is among the configured uris. If not, returns False.

        :param _redirect_uri: URI sent in the Authorization request
        :return:            True if it is among the configured URIs, False otherwise
        """
        part = urlparse(_redirect_uri)
        if part.fragment:
            log.debug("Can not contain fragment: %s", _redirect_uri)
            return False

        _base, _query = self.__split_base_query(_redirect_uri)

        for potential_redirect_uri in self.redirect_uris.split():
            redirect_uri_base, redirect_uri_query = self.__split_base_query(potential_redirect_uri)

            # The base of URI MUST exactly match the base
            if _base != redirect_uri_base:
                continue

            class NotFoundException(Exception):
                pass

            try:
                # every registered query component must exist in the
                # redirect_uri
                if redirect_uri_query:
                    for key, vals in redirect_uri_query.items():
                        if not _query or key not in _query:
                            raise NotFoundException()

                        for val in vals:
                            if val and val not in _query[key]:
                                raise NotFoundException()

                # and vice versa, every query component in the redirect_uri
                # must be registered
                if _query:
                    if redirect_uri_query is None:
                        raise NotFoundException()
                    for key, vals in _query.items():
                        if key not in redirect_uri_query:
                            raise NotFoundException()
                        for val in vals:
                            if redirect_uri_query[key]:
                                if val not in redirect_uri_query[key] and '' not in redirect_uri_query[key]:
                                    raise NotFoundException()
                # found it, so return True
                return True
            except NotFoundException:
                pass

        log.debug("%s Doesn't match any registered uris %s", _redirect_uri, self.redirect_uris)
        return False

    @staticmethod
    def __split_base_query(_redirect_uri):
        (_base, _query) = splitquery(_redirect_uri)
        if _query:
            _query = parse_qs(_query, keep_blank_values=True)
        return _base, _query

    @staticmethod
    def self_instance():
        return OpenIDClient.objects.get(client_id=OpenIDClient.SELF_CLIENT_ID)

    def get_key(self, key_type):
        return OpenIDKey.objects.get(client=self, key_type=key_type).key

    def __str__(self):
        return self.client_name


class OpenIDAgreement(models.Model):
    client = models.ForeignKey(OpenIDClient, on_delete=models.CASCADE, related_name='agreements')
    text = models.TextField()
    obligatory = models.BooleanField()

    allowed_scopes = JSONField(null=True, blank=True,
                               verbose_name='List of allowed scopes for this agreement.')

    allowed_claims = JSONField(null=True, blank=True,
                               verbose_name='List of allowed claims for this agreement')

    username_auto_agreement_regexp = models.CharField(null=True, blank=True, max_length=256,
                                                      verbose_name='Usernames matching this regexp will have this '
                                                                   'agreement automatically agreed to')

    def can_auto_approve(self, username):
        return self.username_auto_agreement_regexp and re.match(self.username_auto_agreement_regexp, username)


class OpenIDUserAgreement(models.Model):
    agreement = models.ForeignKey(OpenIDAgreement, on_delete=models.CASCADE, related_name='user_agreements')
    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name='user_agreements')
    agreed_on = models.DateTimeField()
    agreed_by_user = models.BooleanField(default=False, verbose_name='If set to False, '
                                                                     'the agreement was created automatically, '
                                                                     'otherwise it was created explicitly by user')


class OpenIDToken(models.Model):
    """
        Store for issued tokens. Only the hash is stored, not the token itself.
    """
    client = models.ForeignKey(OpenIDClient, on_delete=models.CASCADE)
    token_hash = models.CharField(max_length=64, unique=True)
    token_type = models.CharField(max_length=4)
    token_data = jsonfield.JSONField(default={})
    expiration = models.DateTimeField()
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    root_token = models.ForeignKey('openid_connect_op.OpenIDToken', related_name='related_tokens',
                                   on_delete=models.PROTECT,
                                   null=True, blank=True)

    @staticmethod
    def get_token_hash(token):
        """
        Returns a sha256 hash for the given token. The hash is used as token_hash attribute

        :param token:   token
        :return:        sha256 hexdigest
        """
        return hashlib.sha256(token.encode('ascii')).hexdigest()

    TOKEN_TYPE_ACCESS_BEARER_TOKEN = 'ACCT'
    TOKEN_TYPE_REFRESH_TOKEN = 'REFR'
    TOKEN_TYPE_ID_TOKEN = 'ID'
    TOKEN_TYPE_CLIENT_DYNAMIC_REGISTRATION = 'CDR'

    @classmethod
    def create_token(cls, client, token_type, token_data, ttl, user, root_db_token=None, token=None):
        """
        Creates a time-limited token of a given type associated with user

        :param client:          for which client is the token created/registered
        :param token_type:      type of the token
        :param token_data:      extra JSON data associated with the token
        :param ttl:             ttl in seconds beginning now
        :param user:            user with whom the token is associated
        :return:                created token as urlsafe string
        """
        if not token:
            token = secrets.token_urlsafe(64)
        token_hash = OpenIDToken.get_token_hash(token)
        db_token = OpenIDToken.objects.create(
            client=client,
            token_hash=token_hash,
            token_type=token_type,
            token_data=token_data,
            expiration=timezone.now() + datetime.timedelta(seconds=ttl)
            if not isinstance(ttl, datetime.datetime) else ttl,
            user=user,
            root_token=root_db_token)

        return token, db_token


class OpenIDKey(models.Model):
    client = models.ForeignKey(OpenIDClient, on_delete=models.CASCADE)

    JWK_RSA_PUBLIC_KEY = 'jwk_rsa_public'
    JWK_RSA_PRIVATE_KEY = 'jwk_rsa_private'
    AES_KEY = 'aes'

    key_type = models.CharField(max_length=16, choices=(
        (JWK_RSA_PRIVATE_KEY, 'JWK RSA private key'),
        (JWK_RSA_PUBLIC_KEY, 'JWK RSA public key'),
        (AES_KEY, 'AES key')
    ))
    encrypted_key_value = models.BinaryField()

    @property
    def key(self):
        return CryptoTools.decrypt(self.encrypted_key_value,
                                   key=settings.OPENID_CONNECT_OP_DB_ENCRYPT_KEY)

    @key.setter
    def key(self, value):
        self.encrypted_key_value = OpenIDKey.encrypt_key(value)

    @staticmethod
    def encrypt_key(value):
        return CryptoTools.encrypt(value, key=settings.OPENID_CONNECT_OP_DB_ENCRYPT_KEY)
